package application1.aspects;

import application1.interfaces.*;
import library.util.*;
import library.interfaces.*;

public aspect PerformanceMonitoringContextAspect 
{
	declare parents: PerformanceMonitoringContext implements
															Collaborative,
															EntrySynchronizing/*,
															Tracing*/;

	public long PerformanceMonitoringContext.totalReadTime = 0;
	public long PerformanceMonitoringContext.totalWriteTime = 0;
	public long PerformanceMonitoringContext.totalUpdateTime = 0;
	
	public int PerformanceMonitoringContext.totalReadCount = 0;
	public int PerformanceMonitoringContext.totalWriteCount = 0;
	public int PerformanceMonitoringContext.totalUpdateCount = 0;

	public long PerformanceMonitoringContext.totalTimeElapsed = 0;
	
	
	// The pointcut which intercepts calls to the application objects originating from application threads
	pointcut accessToTheObject(PerformanceMonitoringObject object, PerformanceMonitoringMember participant):
    	target(object) && this(participant) && call(public * PerformanceMonitoringObject+.*(..));
	
	Object around(PerformanceMonitoringObject object, PerformanceMonitoringMember participant): accessToTheObject(object,participant)
	{
		PerformanceMonitoringContext context = (PerformanceMonitoringContext) participant.getCurrentContext();
		
		long readMethodExecutionStart = 0;
		long writeMethodExecutionStart = 0;
		long updateMethodExecutionStart = 0;
		
		String methodName = (String) thisJoinPoint.getSignature().getName();
		
		if(object.getAccessTypeOfMethod(methodName) == AccessTypes.READ)
		{
			context.totalReadCount++;
			readMethodExecutionStart = System.currentTimeMillis();
		}
		else if(object.getAccessTypeOfMethod(methodName) == AccessTypes.WRITE)
		{
			context.totalWriteCount++;
			writeMethodExecutionStart = System.currentTimeMillis();
		}
		else
		{
			context.totalUpdateCount++;
			updateMethodExecutionStart = System.currentTimeMillis();
		}
			
		Object returningObject = proceed(object,participant);
		
		if(object.getAccessTypeOfMethod(methodName) == AccessTypes.READ)
		{
			context.totalReadTime += System.currentTimeMillis() - readMethodExecutionStart;
		}
		else if(object.getAccessTypeOfMethod(methodName) == AccessTypes.WRITE)
		{
			context.totalWriteTime += System.currentTimeMillis() - writeMethodExecutionStart;
		}
		else
		{
			context.totalUpdateTime += System.currentTimeMillis() - updateMethodExecutionStart;
		}

		return returningObject;
	}
	
	pointcut contextDestruction(PerformanceMonitoringContext context):
		this(context) && call(public * Context.inactivateContext(..));
	
	after(PerformanceMonitoringContext context) : contextDestruction(context)
	{
		context.totalTimeElapsed = System.currentTimeMillis() - context.totalTimeElapsed;
		int totalAccessCount = context.totalReadCount + context.totalWriteCount + context.totalUpdateCount;
		
		System.out.println("\n\n****************************** Statistics / Performance Results *************************************");
		System.out.println("Total Number of Accesses: " + totalAccessCount);
		System.out.println("Total Elapsed Time: " + context.totalTimeElapsed + " milliseconds");
		System.out.println("Efficiency: " + (double) (context.totalTimeElapsed / totalAccessCount) + " milliseconds per access.\n");
		
		System.out.print("The following objects were READ: ");
		for(String objectName: context.getReadObjectNames())
		{
			System.out.print(objectName + ", ");
		}
		System.out.println("\nNumber of Read Accesses: " + context.totalReadCount);
		System.out.println("Elapsed Time For Read Accesses : " + context.totalReadTime + " milliseconds");
		System.out.println("Read Access Efficiency: " + (double)(context.totalReadTime / context.totalReadCount) + " milliseconds per READ access.\n");
		
		System.out.print("The following objects were WRITTEN: ");
		for(String objectName: context.getWrittenObjectNames())
		{
			System.out.print(objectName + ", ");
		}
		System.out.println("\nNumber of Write Accesses: " + context.totalWriteCount);
		System.out.println("Elapsed Time For Write Accesses : " + context.totalWriteTime + " milliseconds");
		System.out.println("Write Access Efficiency: " + (double)(context.totalWriteTime / context.totalWriteCount) + " milliseconds per WRITE access.\n");
		
		
		System.out.print("The following objects were UPDATED: ");
		for(String objectName: context.getUpdatedObjectNames())
		{
			System.out.print(objectName + ", ");
		}
		System.out.println("\nNumber of Update Accesses: " + context.totalUpdateCount);
		System.out.println("Elapsed Time For Update Accesses : " + context.totalUpdateTime + " milliseconds");
		System.out.println("Update Access Efficiency: " + (double)(context.totalUpdateTime / context.totalUpdateCount) + " milliseconds per UPDATE access.");
		
		System.out.println("****************************** Statistics / Performance Results *************************************\n\n");
	}
	
	pointcut contextCreation(PerformanceMonitoringContext context):
    	target(context) && execution(public PerformanceMonitoringContext+.new(..)) ;
	
	after(PerformanceMonitoringContext context) : contextCreation(context)
	{
		context.totalTimeElapsed = System.currentTimeMillis();
	}
}
