package library.aspects.context;

// Allowing the performance of collaborative work.
import library.interfaces.Collaborative;
import library.interfaces.Context;
import library.interfaces.Collaborating;
import library.context.ContextObject;

import java.util.Vector;

import library.util.Debugger;

public aspect CollaborativeAspect introduces Collaborative
{
	//declare parents: Collaborative implements Context;

	private Vector<Collaborating> Collaborative.joinedParticipants;
	private Vector<Collaborating> Collaborative.spawnedParticipants;
	
	private boolean Collaborative.hasBeenClosed = false;
	
	private int Collaborative.maximumMemberCount = 0;
	private boolean Collaborative.maximumNumberSpecified = false;
	
	pointcut constructorExecution(Collaborative newContext):
    	target(newContext) && execution(public Collaborative+.new(..))
    					&& !execution(public ContextObject.new(..));
	
	before(Collaborative newContext) : constructorExecution(newContext)
	{
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,newContext,"Constructor Execution call has been intercepted.",2);
		
		newContext.initializeMembers();
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,newContext,"Collaboration service is initiated.",1);
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,newContext,"Constructor Execution call has been released.",2);
	}
	
	public boolean Collaborative.isClosed()
	{
		return hasBeenClosed;
	}
	
	public void Collaborative.closeContext()
	{
		hasBeenClosed = true;
	}
	
	public void Collaborative.initializeMembers()
	{
		joinedParticipants = new Vector<Collaborating>();
		spawnedParticipants = new Vector<Collaborating>();
	}
	
	public int Collaborative.getParticipantCount()
	{
		return joinedParticipants.size() + spawnedParticipants.size();
	}
	
	public void Collaborative.setMaximumMemberCount(int newMaximum)
	{
		maximumMemberCount = newMaximum;
		maximumNumberSpecified = true;
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"Number of maximum members has been set as : " + newMaximum,1);
	}
	
	public void Collaborative.setMaximumNumberSpecified(boolean value)
	{
		maximumNumberSpecified = value;
	}
	
	public void Collaborative.addParticipant(Collaborating participant)
	{
		if(this.isClosed())
		{
			Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"A participant  -" + Thread.currentThread().getName() + "-" + " has been turned down because the context was closed!",1);
			return;
		}
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The context will add a new participant -" + Thread.currentThread().getName() + "-.",1);
		
		joinedParticipants.add(participant);
		participant.setCurrentContext(this);
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The participant  -" + Thread.currentThread().getName() + "-" + " has now been added to the context.",2);
		
		if(maximumNumberSpecified && getParticipantCount() >= maximumMemberCount)
			hasBeenClosed = true;
	}
	
	public synchronized void Collaborative.addSpawnedParticipant(Collaborating participant)
	{
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The context will add a new spawned participant -" + Thread.currentThread().getName() + "-.",1);
		
		spawnedParticipants.add(participant);
		participant.setCurrentContext(this);
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The spawned participant -" + Thread.currentThread().getName() + "-" + " has now been added to the context.",2);
	}
	
	public synchronized void Collaborative.removeParticipant(Collaborating participant)
	{	
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The context will remove a participant -" + Thread.currentThread().getName() + "-.",1);
		
		// If participant is spawned, remove it from the spawned collection !
		if(!joinedParticipants.remove(participant))
			spawnedParticipants.remove(participant);
		
		participant.setCurrentContext(null);
		
		Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The participant -" + Thread.currentThread().getName() + "-" + " has been removed from the context.",1);
		
		if(joinedParticipants.size() + spawnedParticipants.size() == 0)
		{
			Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The context will now inactivate itself.",1);
			
			inactivateContext();
			
			Debugger.printMessage("Collaborative",Debugger.CONTEXT_ASPECT,this,"The context has been inactivated.",1);
		}
	}
	
	pointcut memberAddition(Collaborative context):
		call(public void assignParticipant(..)) && target(context);
	
	Object around(Collaborative context) : memberAddition(context)
	{
		Collaborating participant = (Collaborating) Thread.currentThread();
		
		context.addParticipant(participant);
		
		return null;
	}
	
	pointcut memberRemoval(Collaborative context):
		call(public void dismissParticipant(..)) && target(context);
	
	Object around(Collaborative context) : memberRemoval(context)
	{
		Collaborating participant = (Collaborating) thisJoinPoint.getArgs()[0];
		
		context.removeParticipant(participant);
		
		return null;
	}
	
	pointcut memberExistanceQuery(Collaborative context):
		call(public boolean hasParticipant(..)) && target(context);
	
	Object around(Collaborative context) : memberExistanceQuery(context)
	{
		Collaborating participant = (Collaborating) thisJoinPoint.getArgs()[0];
		
		return context.containsMember(participant);

	}
	
	pointcut participantNumberQuery(Collaborative context): 	
		target(context)	&& call(public int numberOfParticipants(..));
	
	Object around(Collaborative context) : participantNumberQuery(context)
	{
		return context.getParticipantCount();
	}
	
	
	public synchronized boolean Collaborative.containsMember(Collaborating member)
	{
		if(joinedParticipants.contains(member) || spawnedParticipants.contains(member))
			return true;
		
		return false;
	}
	
}
