package library.aspects.context;

import library.interfaces.Deferring;
import library.interfaces.VersionedContextAware;
import library.interfaces.Tracing;
import library.interfaces.Traceable;
import library.interfaces.Context;
import java.util.Set;
import library.util.AccessTypes;

import library.util.Debugger;

public aspect DeferringAspect 
{
	//declare parents: Deferring implements Tracing;
	
	//Finalization: Make sure that version of the context becomes the main version
	public void Deferring.finalizeVersions()
	{
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,this,"The versions of the context will be finalized.",1);
		
		Set<Object> modifiedObjects = this.getModifiedObjects();
		
		for(Object obj: modifiedObjects)
		{
			VersionedContextAware versioned = (VersionedContextAware) obj;
			
			// First get the current main version
			int oldMainVersion = versioned.findIdOfMainVersion();
			
			// Then find the version of the context (candidate for being the main version)
			int newMainVersion = versioned.lookUpVersionForContext(this);
			
			// Set it as the main
			versioned.setItsMainVersionAs(newMainVersion);
			
			// Delete the old main version
			versioned.deleteVersion(oldMainVersion);
			
			// The context does no longer need to point any specific context
			versioned.unregisterContext(this);
		}
		
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,this,"The finalization of versions has been completed.",2);
	}
	
	pointcut deferringContextCreation(Deferring deferringContext):
		target(deferringContext) && execution(public Deferring+.new(..)) ;
	
	after(Deferring deferringContext) : deferringContextCreation(deferringContext)
	{
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"Constructor Execution (after) call has been intercepted.",2);
		
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"Deferring service inititated.",1);
		
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"Constructor Execution (after) call has been returned.",2);	
	}
	
	pointcut deferringContextDestruction(Deferring deferringContext):
		this(deferringContext) && call(public void Context.inactivateContext(..));
	
	before(Deferring deferringContext) : deferringContextDestruction(deferringContext)
	{
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"Inactivation call (before) has been intercepted.",2);
		
		deferringContext.finalizeVersions();
		
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"Inactivation call (before) has been returned.",2);
	}
	
	pointcut memberPerformingWork(Deferring deferringContext) : 
		call(public * participantPerformsWork(..)) && target(deferringContext);
	
	before(Deferring deferringContext): memberPerformingWork(deferringContext)
	{
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"The notification call (before) has been intercepted.",2);
		
		Object[] allArguments = thisJoinPoint.getArgs();
		
		Object accessedObject = allArguments[1];
		String methodName = (String) allArguments[2];
		
		// We may assume that the object is Traceable since Deferring depends on Tracing
		String accessType = ((Traceable) accessedObject).getAccessTypeOfMethod(methodName);
		String objectName = ((Traceable) accessedObject).getMyName();
		
		if(accessedObject instanceof VersionedContextAware)
		{
			VersionedContextAware versionedObject = (VersionedContextAware) accessedObject;
			
			if(!accessType.equals(AccessTypes.READ))
			{	
				if(!deferringContext.hasModifiedObject(objectName))
				{
					Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"The context is creating a new version of the object.",1);
					
					int newVersionID = versionedObject.createNewVersion();
					
					versionedObject.registerContextForVersion(deferringContext, newVersionID);
				}
				else
				{
					Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"The object was accessed before, therefore no action taken..",2);
				}
			}
			else
			{
				Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"The object will not be modified, therefore no action taken..",2);
			}
		}
		else
		{
			//System.out.println("Object is not VersionedContextAware!");
		}
	
		Debugger.printMessage("Deferring",Debugger.CONTEXT_ASPECT,deferringContext,"The notification call (before) has been released.",2);
	}

}
