package library.aspects.context;

import java.util.Set;
import java.util.LinkedHashSet;

import library.interfaces.Tracing;
import library.interfaces.Traceable;
import library.interfaces.Context;

import library.util.ObjectTrace;
import library.util.AccessTypes;

import library.util.Debugger;

public aspect TracingAspect introduces Tracing
{
	//declare parents: Tracing implements Context;
	
	private Set<ObjectTrace> Tracing.traces;
	
	public void Tracing.initializeTraces()
	{
		traces = new LinkedHashSet<ObjectTrace>();
	}
	
	public void Tracing.finalizeTraces()
	{
		// Regular tracking service does not take any action
	}
	
	public Set<ObjectTrace> Tracing.getAllTraces()
	{
		return traces;
	}
	
	public boolean Tracing.hasReadObject(String objectName)
	{
		Set<ObjectTrace> readTraces = getReadTraces();
		
		for(ObjectTrace track: readTraces)
		{
			if(track.getTrackedObjectName().equals(objectName))
				return true;
		}
		
		return false;
	}
	
	public boolean Tracing.hasReadObject(Object object)
	{
		Traceable trackable = (Traceable) object;
		
		return hasReadObject(trackable.getMyName());
	}
	
	public boolean Tracing.hasModifiedObject(String objectName)
	{
		Set<ObjectTrace> modifiedTraces = getWriteTraces();
		
		modifiedTraces.addAll(getUpdateTraces());
		
		for(ObjectTrace trace: modifiedTraces)
		{
			if(trace.getTrackedObjectName().equals(objectName))
				return true;
		}
		
		return false;
	}
	
	public boolean Tracing.hasModifiedObject(Object object)
	{
		Traceable trackable = (Traceable) object;
		
		return hasModifiedObject(trackable.getMyName());
	}
	
	
	public Set<Object> Tracing.getReadObjects()
	{
		Set<Object> readObjects = new LinkedHashSet<Object>();
		
		for(ObjectTrace trace: traces)
		{
			if(trace.getAccessType().equals(AccessTypes.READ))
				readObjects.add(trace.getTrackedObject());
		}
		
		return readObjects;
	}
	
	public Set<Object> Tracing.getModifiedObjects()
	{
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,this,"The set of modified objects will be provided.",2);
		
		Set<Object> modifiedObjects = new LinkedHashSet<Object>();
		
		for(ObjectTrace trace: traces)
		{
			if(trace.getAccessType().equals(AccessTypes.UPDATE) || 
								trace.getAccessType().equals(AccessTypes.WRITE))
				modifiedObjects.add(trace.getTrackedObject());
		}
		
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,this,"The set of modified objects have been provided.",2);
		
		return modifiedObjects;
	}
	
	public Set<String> Tracing.getReadObjectNames()
	{
		Set<ObjectTrace> readTraces = this.getReadTraces();
		Set<String> readObjectNames = new LinkedHashSet<String>();
		
		for(ObjectTrace trace: readTraces)
		{
			readObjectNames.add(trace.getTrackedObjectName());
		}
		
		return readObjectNames;
	}
	
	public Set<String> Tracing.getWrittenObjectNames()
	{
		Set<ObjectTrace> writeTraces = this.getWriteTraces();
		Set<String> writtenObjectNames = new LinkedHashSet<String>();
		
		for(ObjectTrace trace: writeTraces)
		{
			writtenObjectNames.add(trace.getTrackedObjectName());
		}
		
		return writtenObjectNames;
	}
	
	public Set<String> Tracing.getUpdatedObjectNames()
	{
		Set<ObjectTrace> updateTraces = this.getUpdateTraces();
		Set<String> updatedObjectNames = new LinkedHashSet<String>();
		
		for(ObjectTrace trace: updateTraces)
		{
			updatedObjectNames.add(trace.getTrackedObjectName());
		}
		
		return updatedObjectNames;
	}
	
	public void Tracing.addObjectTrace(ObjectTrace newObjectTrace)
	{
		if(traces.add(newObjectTrace))
		{	
			Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,this,"The new trace has been added [" + newObjectTrace.getAccessType() + 
					" , " + newObjectTrace.getTrackedObjectName() + "].",1);
		}
		
	}
	private Set<ObjectTrace> Tracing.getReadTraces()
	{
		Set<ObjectTrace> readTraces = new LinkedHashSet<ObjectTrace>();
		
		for(ObjectTrace trace : traces)
		{
			if(trace.getAccessType().equals(AccessTypes.READ))
				readTraces.add(trace);
		}
		
		return readTraces;
	}
	
	private Set<ObjectTrace> Tracing.getWriteTraces()
	{
		Set<ObjectTrace> writeTraces = new LinkedHashSet<ObjectTrace>();
		
		for(ObjectTrace trace : traces)
		{
			if(trace.getAccessType().equals(AccessTypes.WRITE))
				writeTraces.add(trace);
		}
		
		return writeTraces;
	}
	
	private Set<ObjectTrace> Tracing.getUpdateTraces()
	{
		Set<ObjectTrace> updateTraces = new LinkedHashSet<ObjectTrace>();
		
		for(ObjectTrace trace : traces)
		{
			if(trace.getAccessType().equals(AccessTypes.UPDATE))
				updateTraces.add(trace);
		}
		
		return updateTraces;
	}
	
	pointcut trackingContextCreation(Tracing trackingContext):
    	target(trackingContext) && execution(public Tracing+.new(..)) ;
	
	after(Tracing trackingContext) : trackingContextCreation(trackingContext)
	{
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Constructor Execution (after) call has been intercepted.",2);
		
		trackingContext.initializeTraces();
		
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Tracing service inititated.",1);
		
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Constructor Execution (after) call has been returned.",2);
		
	}
	
	pointcut trackingContextDestrucion(Tracing trackingContext):
		this(trackingContext) && call(public void Context.inactivateContext(..));
	
	before(Tracing trackingContext) : trackingContextDestrucion(trackingContext)
	{
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Inactivation call (before) has been intercepted.",2);
		
		trackingContext.finalizeTraces();
		
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Inactivation call (before) has been returned.",2);
	}
	
	pointcut memberPerformingWork(Tracing trackingContext) : 
		call(public * participantPerformsWork(..)) && target(trackingContext);
	
	before(Tracing trackingContext): memberPerformingWork(trackingContext)
	{
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"The notification call (before) has been intercepted.",2);
		
		Object[] arguments = thisJoinPoint.getArgs();
		
		Object accessedObject = arguments[1];
		String methodName = (String) arguments[2];
		
		if(accessedObject instanceof Traceable)
		{
			Traceable trackableObject = (Traceable) accessedObject;
			
			Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"Aspect will create a new trace of the accessed object.",1);
			
			ObjectTrace trace = trackableObject.createMyTrace(methodName);
			
			trackingContext.addObjectTrace(trace);
		}
		else
		{
			System.out.println("The object is not Traceable!");
		}
		
		Debugger.printMessage("Tracing",Debugger.CONTEXT_ASPECT,trackingContext,"The notification call (before) has been released.",2);
	}
	
	
}
