package library.aspects.context;

import library.interfaces.TwoPhaseLocking;
import library.interfaces.Tracing;
import library.interfaces.Traceable;
import library.interfaces.Context;
import library.interfaces.Lockable;
import java.util.Set;
import library.util.AccessTypes;

import library.util.Debugger;

public aspect TwoPhaseLockingAspect introduces TwoPhaseLocking 
{
	//declare parents: TwoPhaseLocking implements Tracing;

	public boolean TwoPhaseLocking.hasAcquiredReadLockOfObject(Lockable queriedObject)
	{
		return hasReadObject(queriedObject);
	}
	
	public boolean TwoPhaseLocking.hasAcquiredWriteLockOfObject(Lockable queriedObject)
	{
		return hasModifiedObject(queriedObject);
	}
	
	public boolean TwoPhaseLocking.hasAcquiredLockOfObject(Lockable queriedObject)
	{
		if(hasAcquiredReadLockOfObject(queriedObject) || hasAcquiredWriteLockOfObject(queriedObject))
			return true;
		else
			return false;
	}
	
	public void TwoPhaseLocking.releaseAllAcquiredLocks()
	{
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"All the locks of the context are being released.",1);
		
		Set<Object> readObjects = getReadObjects();
		Set<Object> modifiedObjects = getModifiedObjects();
		
		for(Object object : readObjects)
		{
			Lockable lockableObject = (Lockable) object;
			
			// The lock could have been ugraded, therefore we need to check it
			if(!hasAcquiredWriteLockOfObject(lockableObject))
			{		
				Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The READ lock of an object is being released.",1);
				
				lockableObject.releaseReadLock("2PL");
				
				Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The READ lock of an object has been released.",2);
			}
		}
		
		for(Object object : modifiedObjects)
		{
			Lockable lockableObject = (Lockable) object;
			
			Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The WRITE lock of an object is being released.",1);
			
			lockableObject.releaseWriteLock("2PL");
			
			Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The WRITE lock of an object has been released.",2);
			
		}
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"All the locks of the context have been released.",2);
	}
	

	
	public void TwoPhaseLocking.upgradeTheAcquiredReadLockOfObject(Lockable lockableObject)
	{
		// Atomic upgrade operation
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The lock of an object will be upgraded.",1);
		
		lockableObject.upgradeLock("2PL");
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The lock of an object has been upgraded.",2);
		// Atomic upgrade operation
	}
	
	public void TwoPhaseLocking.acquireNecessaryLock(Object accessedObject, String methodName)
	{
		if(accessedObject instanceof Lockable)
		{
			Lockable lockableObject = (Lockable) accessedObject;
			
			String accessType = ((Traceable)lockableObject).getAccessTypeOfMethod(methodName);
			
			if(accessType.equals(AccessTypes.READ))
			{
				if(!hasAcquiredLockOfObject(lockableObject))
				{
					Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The READ lock of an object is about to be acquired.",1);
					
					lockableObject.acquireReadLock("2PL");
					
					Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The READ lock of an object has been acquired.",2);
				}
					
			}
			
			else if(accessType.equals(AccessTypes.WRITE) || accessType.equals(AccessTypes.UPDATE))
			{
				if(hasAcquiredLockOfObject(lockableObject))
				{
					// Had read lock, now requesting a write lock.. Upgrade operation will take place
					if(hasAcquiredReadLockOfObject(lockableObject))
					{
						upgradeTheAcquiredReadLockOfObject(lockableObject);
					}
					
					// Has already a write lock.. Do nothing
					else
					{
						
					}
					
				}
				
				// Doesn't have any lock.. Write lock will be acquired..
				else
				{
					Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The WRITE lock of an object is about to be acquired.",1);
					
					lockableObject.acquireWriteLock("2PL");
		
					Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,this,"The WRITE lock of an object has been acquired.",2);
				}
			}
		}
		else
		{
			//System.out.println("The object is not lockable!");
		}
	}
	
	pointcut lockingContextCreation(TwoPhaseLocking lockingContext):
    	target(lockingContext) && execution(public TwoPhaseLocking+.new(..)) ;
	
	after(TwoPhaseLocking lockingContext) : lockingContextCreation(lockingContext)
	{	
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"Constructor Execution call (after) has been intercepted.",2);
		
		//lockingContext.initializeTracks();
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"TwoPhaseLocking service inititated.",1);
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"Constructor Execution call (after) has been returned.",2);
	}
	
	pointcut memberPerformingWork(TwoPhaseLocking twoPLContext) : 
		call(public * participantPerformsWork(..)) && target(twoPLContext);
	
	before(TwoPhaseLocking lockingContext): memberPerformingWork(lockingContext)
	{
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"The notification call (before) has been intercepted.",2);
		
		Object[] arguments = thisJoinPoint.getArgs();
		
		Object accessedObject = arguments[1];
		String methodName = (String) arguments[2];
		
		lockingContext.acquireNecessaryLock(accessedObject,methodName);
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"The notification call (before) has been released.",2);
	}
	
	pointcut twoPLContextDestrucion(TwoPhaseLocking twoPLContext):
		this(twoPLContext) && call(public void Context.inactivateContext(..));
	
	after(TwoPhaseLocking lockingContext) : twoPLContextDestrucion(lockingContext)
	{	
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"Inactivation call (after) has been intercepted.",2);
		
		lockingContext.releaseAllAcquiredLocks();
		
		Debugger.printMessage("TwoPhaseLocking",Debugger.CONTEXT_ASPECT,lockingContext,"Inactivation Execution call (after) has been returned.",2);
	}
	
}

