package library.aspects.object;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import java.util.*;

import library.interfaces.Versioned;
import library.interfaces.Copyable;
import library.util.Debugger;

/** AspectJ Limitation **/
import library.interfaces.VersionedContextAware;
/** AspectJ Limitation **/

public aspect VersionedAspect introduces Versioned
{
	//declare parents: Versioned implements Copyable;

	// Keeps a reference to all the versions of the instance
	// There will only be 1 copy of this vector
	private Vector<Versioned> Versioned.myVersions = new Vector<Versioned>();
	private Map<Object,Integer> Versioned.mappings = new LinkedHashMap<Object,Integer>();
	
	private int Versioned.versionIDGenerator = 0;
	private boolean Versioned.mainVersion;  
	private int Versioned.versionID = 0;
    
	pointcut versionedConstructorExecution(Versioned versionedObject):
    	target(versionedObject) && execution(public Versioned+.new(..));
	
	after(Versioned versionedObject) : versionedConstructorExecution(versionedObject)
	{
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"Constructor Execution (after) call has been intercepted.",2);
		
		versionedObject.setVersionID(0);
    	versionedObject.setItselfAsTheMainVersion();
    	
    	Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"Versioned object has been assigned the version -0-.",1);
    	
    	versionedObject.addNewVersion(versionedObject);
    	
    	Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"Constructor Execution (after) call has been returned.\n",2);
	}
	
	pointcut anyMethodCall(Versioned versionedObject): 
		target(versionedObject) && call(public * Versioned+.*(..)) && 
			!call(public * Versioned.*(..)) && !call(public * VersionedContextAware.*(..));
			
	Object around(Versioned versionedObject) : anyMethodCall(versionedObject)
	{
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"The method call (around) [" + thisJoinPoint.getSignature().getName() + 
				"] to the object has been intercepted.",2);
		
		Object returningObject = null;
		
		int correctVersionToInvoke = versionedObject.findCorrectVersionToInvoke();
		
		if(versionedObject.getVersionID() == correctVersionToInvoke)
		{
			Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"The call will proceed so that the invocation takes place" +
					" on version -" + correctVersionToInvoke + "-.",2);
			
			returningObject =  proceed(versionedObject);
			
			Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"The call has been re-intercepted by the aspect.",2);
		}
		
		else
		{
			String methodName = thisJoinPoint.getSignature().getName();
			Object args[] = thisJoinPoint.getArgs();
			
			Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"It is made sure that the invocation reflectively takes place" +
					" on version -" + correctVersionToInvoke + "-.",1);
			
			returningObject = versionedObject.invokeMethodOnSomeVersion(correctVersionToInvoke,methodName,args);
			
			Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"The call has been re-intercepted by the aspect.",2);
		}
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,versionedObject,"The method call (around) [" + thisJoinPoint.getSignature().getName() + 
				"] to the object has been returned.\n",2);	
		
		return returningObject;
	}
	
	public Vector<Versioned> Versioned.getMyVersions()
	{
		return myVersions;
	}
	
	public void Versioned.setMyVersions(Vector<Versioned> newVersions)
	{
		myVersions = newVersions;
	}
	
	public Map<Object,Integer> Versioned.getAllMappings()
	{
		return mappings;
	}
	
	public void Versioned.setMappings(Map<Object,Integer> newMap)
	{
		mappings = newMap;
	}
	
	public void Versioned.setVersionID(int id)
    { 
    	versionID = id; 
    } 
    
	public int Versioned.getVersionID()
    { 
    	return versionID; 
    }
    
	public boolean Versioned.isMainVersion()
    { 
    	return mainVersion; 
    }
    
    public void Versioned.setItselfAsNonMainVersion()
    {
    	mainVersion = false;
    }
    
    public int Versioned.findCorrectVersionToInvoke()
    {
    	// take a look if the thread had registered itself to a specific version
    	if(mappings.containsKey(Thread.currentThread()))
    		return mappings.get(Thread.currentThread());
    	
    	// thread doesn't have a special view.. simply forward to the main version
    	else
    		return this.findIdOfMainVersion();
    }
    
    /** Loop through all the versioned objects and make sure that none of these versions are the main version.
     *  Finally, make yourself the main version.
     */
	public void Versioned.setItselfAsTheMainVersion() 
	{ 	
		for(Versioned object : myVersions)
		{
			object.setItselfAsNonMainVersion();
		}
		 
		 this.mainVersion = true;
	}
	
	/** Given the version, it finds the object with that version, and makes that object
	 * as the main version.
     */
	public void Versioned.setItsMainVersionAs(int versionID)
	{
		for(Versioned object : myVersions)
		{
			if(object.getVersionID() == versionID)
				object.setItselfAsTheMainVersion();
		}
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The main version of the object is now: -" + versionID + "-.",1);
	}
	
	public int Versioned.findIdOfMainVersion()
	{
		for(Versioned object : myVersions)
		{
			if(object.isMainVersion())
				return object.getVersionID();
		}
		
		return -1;
	}
	
	public synchronized int Versioned.createNewVersion()
	{
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"A new version is being created.",1);
		
		this.versionIDGenerator++;
		
		Versioned versionedObject = (Versioned) this.cloneState();
		
		versionedObject.setVersionID(versionIDGenerator);
		versionedObject.setItselfAsNonMainVersion();
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The object has now a new version: -" + versionIDGenerator + "-.",2);
		
		this.myVersions.add(versionedObject);
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The process of creating a new version has ended.",2);
		
		return versionedObject.getVersionID();
	}
	
	public void Versioned.deleteVersion(int id) 
	{ 
		Versioned toBeDeleted = null;
		
		for(Versioned object: myVersions)
		{
			if(object.getVersionID() == id)
			{
				toBeDeleted = object;
			}
		}
		
		this.myVersions.remove(toBeDeleted);
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The object has deleted its version: -" + id + "-.",1);
	} 
	
	public void Versioned.addNewVersion(Versioned newObject)
	{
		this.myVersions.add(newObject);
	}
	
	public Versioned Versioned.getVersionedObjectWithID(int id)
	{
		for(Versioned object : myVersions)
		{
			if(object.getVersionID() == id)
				return object;
		}
		
		return null;
	}
	
	public void Versioned.registerThreadForVersion(int versionID)
	{
		mappings.put(Thread.currentThread(),versionID);
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The thread [" +  Thread.currentThread().getName() + "]" +
				" has been registered to version: -" + versionID + "-.",1);
	}
	
	public void Versioned.unregisterThread()
	{
		mappings.remove(Thread.currentThread());
		
		Debugger.printMessage("Versioned",Debugger.OBJECT_ASPECT,this,"The thread [" +  Thread.currentThread().getName() + "]" +
				" has been unregistered from version: -" + versionID + "-.",1);
	}
	
	public synchronized Object Versioned.invokeMethodOnMainVersion(String methodName, Object[] parameters)						
	{
		Object targetObject = null;
		Object returnedObject = null;	
		
		for(Versioned object : myVersions)
		{
			if(object.isMainVersion())
				targetObject = object;
		}
		
		returnedObject = invokeMethodByReflection(targetObject,parameters,methodName);
		
		return returnedObject;
	}
	
	
	public synchronized Object Versioned.invokeMethodOnSomeVersion(int objID, String methodName, Object[] parameters)						
	{
		Object targetObject = null;
		Object returnedObject = null;
		
		for(Versioned object : myVersions)
		{
			if(object.getVersionID() == objID)
				targetObject = object;
		}
		
		returnedObject = invokeMethodByReflection(targetObject,parameters,methodName);
		
		return returnedObject;
	}	 
	
	private static Object invokeMethodByReflection(Object targetObject,Object[] parameters,String methodName)
	{		
		Object returnedObject = null;
		
		for(Method m: targetObject.getClass().getMethods())
		{
			if(m.getName().equals(methodName))
			{
				try
				{
					returnedObject = m.invoke(targetObject,parameters);
				}
				catch(IllegalAccessException e)
				{
					e.printStackTrace();
				}
				catch(IllegalArgumentException e)
				{
					e.printStackTrace();
				}
				catch(InvocationTargetException e)
				{
					e.printStackTrace();
				}
				
				break;
			}
		}
	
		return returnedObject;
	}
}
